/*	The purpose of this exercise is, to demonstrate a hypothetical way,
 * in which a series of 24-bit PCM-encoded samples, could be ecnoded
 * to a series of DSD-encoded, 1-bit samples.
 *
 * This exercise assumes a low-pass filter, as well as a
 * 16-bit dithering pattern.
 *
 *  The concept of shaped dithering noise will be, that a
 * 16-bit sequence is to be counted linearly, the bit-order reversed,
 * and an offset applied, to minimize the DC component of the
 * dithering pattern. The resulting dithering pattern
 * will be added to the 16 LSBs of the local 24-bit PCM values, before
 * they are compared with the input samples.
 *
 * 	The design of the shaped noise has been kept this simple, based
 * on the observation, that when a regular saw-wave seems to result
 * from a linear progression of integers, the least-significant bit
 * is oscillating at the Nyquist Frequency, the bit to the left of
 * that at half Nyquist Frequency, the next bit at quarter
 * Nyquist Frequency, etc. This suggest that all the powers of two
 * times the frequency of the saw wave, are already frequency
 * components, but with half amplitude each time. This is the reverse
 * of how the dithering noise should be shaped.
 *
 * 	Reversing the bit-order will ensure that the highest amplitude
 * in the dithering noise will be at Nyquist Frequency, and that
 * every time its frequencies are halved, the corresponding amplitudes
 * will also be halved.
 *
 * 	This description of the dithering noise based on the dithering
 * pattern is only approximate, since the waveform with which
 * each bit is oscillating, is also a square wave and not a
 * sine-wave, multiplied by the significance of the bit. This last
 * observation suggests, that for each power of 1/2, there will also
 * be odd harmonics. Which in turn accentuates, by how much
 * higher frequency-components will have greater energy, than the lower
 * frequency-components.
 *
 */

#include "DSD_Encode.h"

encoder_state enc_state_1 = {0, 0};
encoder_state enc_state_2 = {0, 0};

int32_t dither(encoder_state *state)
{
    int32_t temp = 0;
    int32_t dither_count_temp = 0;

    dither_count_temp = state->dither_count + 1;
    if (dither_count_temp > 0xffff)
        dither_count_temp = 0;

    state->dither_count = dither_count_temp;

    for (int i = 0; i < 16; i++) {
        temp = (temp << 1) + (dither_count_temp & 1);
        dither_count_temp = dither_count_temp >> 1;
    }

    return temp - 0x8000;
}

//  The possible domain of D.C. input values will be:
//  +/- 01111111 10000000 00000000
//  (Without causing distortion.)
//  However, when decoded, the resulting D.C. values will be
//  in the range of:
//  +/- 01111111 00000000 00000000
//  For frequencies higher than zero, the LP filter
//  will additionally need to be applied to the allowable range of
//  input values.

int dsd_encode(int32_t pcm_sample, encoder_state *state)
{
    int out_flag = 0;
    int32_t temp = 0;

    temp = state->local_mod_lp;

    if (pcm_sample > temp + dither(state))
        out_flag = 1;

    temp = out_flag ? temp + 0x10000 : temp - 0x10000;

    //  For DSD64, the corner frequency of the LP filter will be
    //  3.537kHz
    //  The decoder must also apply this, and
    //  Frequencies considerably higher, will need to be of reduced
    //  amplitude, to avoid slewing.

    //  The fact should be understood that, as long as the
    //  LP filter used in the decoder matches the one in the
    //  encoder, the actual frequency response will remain uniform 'over
    //  the entire spectrum'.

    temp *= 127;
    state->local_mod_lp = temp >> 7;

    return out_flag;
}
